@@ -73,6 +73,19 @@ static bool allowsBearerTokenRetrieve(const Davix::Uri& uri)
            (uri.getProtocol().rfind("davs", 0) == 0);
 }
 
+bool GfalHttpPluginData::writeFlagFromOperation(const OP& operation) {
+    return (operation == OP::MKCOL) || (operation == OP::WRITE) ||
+           (operation == OP::WRITE_PASV);
+}
+
+bool GfalHttpPluginData::searchFlagFromOperation(const OP& operation) {
+    return (operation == OP::MKCOL) || (operation == OP::HEAD);
+}
+
+bool GfalHttpPluginData::needsTransferHeader(const OP& operation) {
+    return (operation == OP::READ_PASV) || (operation == OP::WRITE_PASV);
+}
+
 static bool isS3SignedURL(const Davix::Uri& url)
 {
     return (url.queryParamExists("AWSAccessKeyId") && url.queryParamExists("Signature")) ||
@@ -104,101 +117,142 @@ static bool gfal_http_get_x509_cert_pair(gfal2_context_t handle, const Davix::Ur
     return certificate_pair;
 }
 
-char* GfalHttpPluginData::retrieve_se_token(Davix::RequestParams& params, const Davix::Uri& uri,
-                                            bool write_access, unsigned validity)
+char* GfalHttpPluginData::find_se_token(const Davix::Uri& uri, const OP& operation)
+{
+    using credTuple = std::pair<std::string, std::string>;
+
+    if (!allowsBearerTokenRetrieve(uri)) {
+        return NULL;
+    }
+
+    bool write_access = writeFlagFromOperation(operation);
+    bool extended_search = searchFlagFromOperation(operation);
+
+    // Helper function to find a token in the Gfal HTTP internal token map
+    auto find_in_token_map = [&](const char* token, const char* token_path, bool write_access) -> bool {
+        auto it = token_map.find(token);
+
+        if (it == token_map.end()) {
+            gfal2_log(G_LOG_LEVEL_DEBUG,
+                      "(SEToken) Retrieved token not in token access map (path=%s) (assuming user-set)",
+                      token_path);
+            return true;
+        }
+
+        if (it->second || (write_access == it->second)) {
+            gfal2_log(G_LOG_LEVEL_DEBUG, "(SEToken) Found token in credential_map[%s] (access=%s) (needed=%s)",
+                      token_path, it->second ? "write" : "read", write_access ? "write" : "read");
+            return true;
+        }
+
+        return false;
+    };
+
+    // Helper function to extract all credentials of type "BEARER"
+    auto cred_map_callback = [](const char* url_prefix, const gfal2_cred_t* cred, void* user_data) {
+        auto cred_list = static_cast<std::list<credTuple>*>(user_data);
+
+        if (strcmp(cred->type, GFAL_CRED_BEARER) == 0) {
+            cred_list->emplace_back(url_prefix, cred->value);
+        }
+    };
+
+    // Helper function to check whether a path contains the search path
+    auto path_contains = [](const std::string& path, const std::string& search) -> bool {
+        size_t pos = path.find(search);
+
+        if (pos == 0) {
+            return (path.size() == search.size()) ||
+                   (path.at(search.size() - 1) == '/') ||
+                   (path.at(search.size()) == '/');
+        }
+
+        return false;
+    };
+
+    std::list<credTuple> cred_list;
+    gfal2_cred_foreach(handle, cred_map_callback, &cred_list);
+
+    for (auto cred = cred_list.begin(); cred != cred_list.end(); cred++) {
+        if (path_contains(uri.getString(), cred->first) ||
+            (extended_search && path_contains(cred->first, uri.getString()))) {
+            if (find_in_token_map(cred->second.c_str(), cred->first.c_str(), write_access)) {
+                return g_strdup(cred->second.c_str());
+            }
+        }
+    }
+
+    // Search token for the full host (backwards compatibility with FTS)
+    GError* error = NULL;
+    char* token = gfal2_cred_get(handle, GFAL_CRED_BEARER, uri.getHost().c_str(), NULL, &error);
+    g_clear_error(&error);
+
+    return token;
+}
+
+char* GfalHttpPluginData::retrieve_and_store_se_token(const Davix::Uri& uri, const OP& operation, unsigned validity)
 {
     bool retrieve_token = gfal2_get_opt_boolean_with_default(handle, "HTTP PLUGIN", "RETRIEVE_BEARER_TOKEN", false);
+    GError* error = NULL;
+    char* token = NULL;
 
     if (!retrieve_token || !allowsBearerTokenRetrieve(uri)) {
         return NULL;
     }
 
+    Davix::RequestParams params = reference_params;
+    get_params_internal(params, uri);
+    get_certificate(params, uri);
+
+    bool write_access = writeFlagFromOperation(operation);
     TokenRetriever* retriever = token_retriever_chain.get();
+
     while (retriever != NULL) {
         try {
             gfal_http_token_t http_token = retriever->retrieve_token(uri, params, write_access, validity);
-            char *token = strdup(http_token.token.c_str());
-            return token;
+            token = strdup(http_token.token.c_str());
+            break;
         } catch (const Gfal::CoreException& e) {
             gfal2_log(G_LOG_LEVEL_INFO, "(SEToken) Error during token retrieval: %s", e.what());
             retriever = retriever->next();
         }
     }
 
-    gfal2_log(G_LOG_LEVEL_WARNING, "(SEToken) Could not retrieve any token for %s", uri.getString().c_str());
-    return NULL;
+    if (!token) {
+        gfal2_log(G_LOG_LEVEL_WARNING, "(SEToken) Could not retrieve any token for %s", uri.getString().c_str());
+        return NULL;
+    }
+
+    // Tokens are treated as opaque, therefor they are cached in the TokenAccessMap
+    // together with write access and validity info
+    gfal2_cred_t* token_cred = gfal2_cred_new(GFAL_CRED_BEARER, token);
+
+    if (gfal2_cred_set(handle, uri.getString().c_str(), token_cred, &error) < 0) {
+        gfal2_log(G_LOG_LEVEL_DEBUG, "(SEToken) Failed to set bearer token in credential_map[%s] due to error: %s",
+                  uri.getString().c_str(), error->message);
+        g_clear_error(&error);
+    } else {
+        gfal2_log(G_LOG_LEVEL_DEBUG, "(SEToken) Set bearer token in credential_map[%s] (access=%s) (validity=%u)",
+                  uri.getString().c_str(), write_access ? "write" : "read" , validity);
+        token_map[token] = write_access;
+    }
+
+    gfal2_cred_free(token_cred);
+    return token;
 }
 
 bool GfalHttpPluginData::get_token(Davix::RequestParams& params, const Davix::Uri& uri,
-                                   bool write_access, unsigned validity,
-                                   bool secondary_endpoint)
+                                   const OP& operation, unsigned validity)
 {
     if (isS3SignedURL(uri)) {
 	    return false;
     }
 
-    GError* error = NULL;
-    const char* token_path = NULL;
-    gchar* token = gfal2_cred_get(handle, GFAL_CRED_BEARER,
-                                  uri.getString().c_str(),
-                                  &token_path, &error);
-    g_clear_error(&error);
+    gchar* token = find_se_token(uri, operation);
 
     if (!token) {
-        // If we don't have specific token for requested URL fallback to token stored for hostname.
-        // TPC with macaroon tokens needs at least BEARER set for full source URL
-        // and hostname BEARER for destination (needed to create missing parent directories)
-        token = gfal2_cred_get(handle, GFAL_CRED_BEARER,
-                               uri.getHost().c_str(),
-                               &token_path, &error);
-        g_clear_error(&error);
-    }
-
-    if (token) {
-        // Check token read/write access in TokenAccessMap
-        // TODO: Replace with inspection of access and validity in the future
-        TokenAccessMap::iterator it = token_map.find(token);
-        if (it != token_map.end()) {
-            gfal2_log(G_LOG_LEVEL_DEBUG, "(SEToken) Found token in credential_map[%s] (access=%s) (needed=%s)",
-                      token_path, it->second ? "write" : "read", write_access ? "write" : "read");
-
-            if (write_access && it->second != write_access) {
-                gfal2_log(G_LOG_LEVEL_INFO, "(SEToken) Invalidating token for path=%s because write access is missing",
-                          token_path);
-                gfal2_cred_del(handle, GFAL_CRED_BEARER, token_path, &error);
-                token_map.erase(it);
-                g_free(token);
-                token = NULL;
-            }
-        } else {
-            gfal2_log(G_LOG_LEVEL_DEBUG, "(SEToken) Retrieved token not in token access map (path=%s) (assuming user-set)",
-                      token_path);
-        }
-    }
-
-    if (!token) {
-        // No token for neither the URL or the hostname
         // Attempt to obtain SE-issued token
-        token = retrieve_se_token(params, uri, write_access, validity);
-
-        if (token) {
-            // Treat tokens as opaque. To handle validity and write access,
-            // cache the token in the TokenAccessMap together with write access and validity info
-
-            // Register token in the credentials map
-            gfal2_cred_t* token_cred = gfal2_cred_new(GFAL_CRED_BEARER, token);
-            if (gfal2_cred_set(handle, uri.getString().c_str(), token_cred, &error) < 0) {
-                gfal2_log(G_LOG_LEVEL_DEBUG, "(SEToken) Failed to set bearer token in credential_map[%s] due to error: %s",
-                          uri.getString().c_str(), error->message);
-                g_clear_error(&error);
-            } else {
-                gfal2_log(G_LOG_LEVEL_DEBUG, "(SEToken) Set bearer token in credential_map[%s] (access=%s) (validity=%u)",
-                          uri.getString().c_str(), write_access ? "write" : "read" , validity);
-                token_map[token] = write_access;
-            }
-
-            gfal2_cred_free(token_cred);
-        }
+        token = retrieve_and_store_se_token(uri, operation, validity);
     }
 
     if (!token) {
@@ -209,9 +263,9 @@ bool GfalHttpPluginData::get_token(Davix::RequestParams& params, const Davix::Ur
     ss << "Bearer " << token;
 
     gfal2_log(G_LOG_LEVEL_INFO, "Using bearer token for HTTPS request authorization%s",
-              secondary_endpoint ? " (passive TPC)" : "");
+              needsTransferHeader(operation) ? " (passive TPC)" : "");
 
-    if (secondary_endpoint) {
+    if (needsTransferHeader(operation)) {
         params.addHeader("TransferHeaderAuthorization", ss.str());
         // Disable credential delegation if we have a valid token for the destination
         params.addHeader("Credential", "none");
@@ -411,22 +465,22 @@ void GfalHttpPluginData::get_gcloud_credentials(Davix::RequestParams& params, co
     g_free(gcloud_json_string);
 }
 
-void GfalHttpPluginData::get_reva_credentials(Davix::RequestParams &params, const Davix::Uri &uri, bool token_write_access)
+void GfalHttpPluginData::get_reva_credentials(Davix::RequestParams &params, const Davix::Uri &uri, const OP& operation)
 {   
     // The authentication with Reva is yet to be properly designed.
     // Nevertheless, there is a need to associate tokens to URLs given that a token issued for a destination must have write rights,
     // whereas right now GFAL issues a `stat()` == `HEAD` request also to the destination without requiring write rights.
     // To be further developed.
 
+    bool write_access = writeFlagFromOperation(operation);
     reva::CredentialProvider provider;
     reva::Credentials creds = params.getRevaCredentials();
-    provider.updateCredentials(creds, uri.getString(), token_write_access);
+    provider.updateCredentials(creds, uri.getString(), write_access);
     params.setRevaCredentials(creds);
 }
 
 void GfalHttpPluginData::get_credentials(Davix::RequestParams& params, const Davix::Uri& uri,
-                                         bool token_write_access, unsigned token_validity,
-                                         bool secondary_endpoint)
+                                         const OP& operation, unsigned token_validity)
 {
     // Setup GSI in case the storage endpoint tries to fall back to GridSite delegation.
     // That does mean that we might contact the endpoint with both X509 and token auth,
@@ -441,11 +495,10 @@ void GfalHttpPluginData::get_credentials(Davix::RequestParams& params, const Dav
     } else if (uri.getProtocol().compare(0, 5, "swift") == 0) {
         get_swift_params(params, uri);
     }else if (uri.getProtocol().compare(0, 3, "cs3") == 0) {
-        get_reva_credentials(params, uri, token_write_access);
-    }// Use bearer token (other authentication mechanism should be disabled)
+        get_reva_credentials(params, uri, operation);
+    } // Use bearer token (other authentication mechanism should be disabled)
       // Not the case for the moment, as certificates are still used (but should be unset in the future)
-    else if (!get_token(params, uri, token_write_access, token_validity,
-                        secondary_endpoint)) {
+    else if (!get_token(params, uri, operation, token_validity)) {
         // Utilize AWS or GCLOUD tokens if no bearer token is available (to be reviewed)
         get_aws_params(params, uri);
         get_gcloud_credentials(params,uri);
@@ -546,13 +599,13 @@ void GfalHttpPluginData::get_tpc_params(Davix::RequestParams* req_params,
 
     if (push_mode) {
         get_params_internal(*req_params, src_uri);
-        get_credentials(*req_params, src_uri, false, token_timeout);
-        get_credentials(*req_params, dst_uri, true, token_timeout, true);
+        get_credentials(*req_params, src_uri, OP::READ, token_timeout);
+        get_credentials(*req_params, dst_uri, OP::WRITE_PASV, token_timeout);
         do_delegation = delegation_required(dst_uri);
     } else {  // Pull mode
         get_params_internal(*req_params, dst_uri);
-        get_credentials(*req_params, src_uri, false, token_timeout, true);
-        get_credentials(*req_params, dst_uri, true, token_timeout);
+        get_credentials(*req_params, src_uri, OP::READ_PASV, token_timeout);
+        get_credentials(*req_params, dst_uri, OP::WRITE, token_timeout);
         do_delegation = delegation_required(src_uri);
     }
 
@@ -582,14 +635,13 @@ void GfalHttpPluginData::get_tpc_params(Davix::RequestParams* req_params,
 
 }
 
-void GfalHttpPluginData::get_params(Davix::RequestParams* req_params,
-                                    const Davix::Uri& uri,
-                                    bool token_write_access)
+void GfalHttpPluginData::get_params(Davix::RequestParams* req_params, const Davix::Uri& uri,
+                                    const OP& operation)
 {
     *req_params = reference_params;
 
     get_params_internal(*req_params, uri);
-    get_credentials(*req_params, uri, token_write_access);
+    get_credentials(*req_params, uri, operation);
 }
 
 
